import lxml.etree as ET
from . import user_namespace as un
from . import utilities as util

# Graph of a 1-variable function
# We'll set up an SVG path element by sampling the graph
# on an equally spaced mesh
def graph(element, diagram, parent, outline_status = None):
    # if we've already added an outline, just plot the graph
    if outline_status == 'finish_outline':
        finish_outline(element, diagram, parent)
        return

    # by default, the domain is the width of the bounding box
    bbox = diagram.bbox()
    domain = element.get('domain')
    if domain is None:
        domain = [bbox[0], bbox[2]]
    else:
        domain = un.valid_eval(domain)

    # retrieve the function from the namespace and generate points
    f = un.valid_eval(element.get('function'))
    N = int(element.get('N', 100))
    dx = (domain[1] - domain[0])/N
    x = domain[0]
    cmds = []
    for _ in range(N+1):
        p = diagram.transform((x, f(x)))
        if len(cmds) == 0:
            cmds = ['M', util.pt2str(p)]
        else:
            cmds += ['L', util.pt2str(p)]
        x += dx

    # now set up the attributes
    util.set_attr(element, 'thickness', '2')
    util.set_attr(element, 'stroke', 'blue')
    if diagram.output_format() == 'tactile':
        element.set('stroke', 'black')

    attrib = {'id': diagram.find_id(element, element.get('id'))}
    attrib.update(util.get_1d_attr(element))
    attrib.update(
        {
            'd': ' '.join(cmds),
            'fill': 'none',
            'type': 'function-graph'
        }
    )

    path = ET.Element('path', attrib = attrib)

    # By default, we clip the graph to the bounding box
    if element.get('cliptobbox') is None:
        element.set('cliptobbox', 'yes')
    util.cliptobbox(path, element, diagram)

    # Finish up handling any requested outlines
    if outline_status == 'add_outline':
        diagram.add_outline(element, path, parent)
        return

    if element.get('outline', 'no') == 'yes' or diagram.output_format() == 'tactile':
        diagram.add_outline(element, path, parent)
        finish_outline(element, diagram, parent)
    else:
        parent.append(path)

def finish_outline(element, diagram, parent):
    diagram.finish_outline(element,
                           element.get('stroke'),
                           element.get('thickness'),
                           element.get('fill', 'none'),
                           parent)





from . import (
    annotations,
    area,
    arrow,
    calculus,
    circle,
    clip,
    coordinates,
    CTM,
    definition,
    diagram,
    diffeqs,
    graph,
    grid_axes,
    group,
    implicit,
    label,
    line,
    math_utilities,
    network,
    parametric_curve,
    parse,
    path,
    point,
    polygon,
    rectangle,
    repeat,
    riemann_sum,
    shape,
    slope_field,
    tags,
    tangent_line,
    user_namespace,
    utilities,
    vector
)





import lxml.etree as ET
import networkx as nx
import numpy as np
import math
import copy
from . import user_namespace as un
from . import math_utilities as math_util
from . import utilities as util
from . import coordinates
from . import label
from . import CTM
from . import group

# Add a graphical element describing a network
def network(element, diagram, parent, outline_status):
    if outline_status == 'finish_outline':
        coords = diagram.get_network_coordinates(element)
        coordinates.coordinates(coords, diagram, parent, outline_status)
        return

    # Is the network directed?
    directed = element.get('directed', 'no') == 'yes'

    # Let's see if there is a dictionary defining the graph
    graph = element.get('graph', None)
    graph_dict = {}
    if graph is not None:
        graph = un.valid_eval(graph)
        if not isinstance(graph, dict):
            print("@graph attribute of a <network> element should be a dictionary")
            return
        for key, value in graph.items():
            value = [str(v) for v in value]
            graph_dict[str(key)] = value
        
    # We'll keep track of the number of edges in a multigraph
    loop_count = {}      # dictionary for number of loops
    directed_edges = {}  # dictionary to count directed edges
    all_edges = {}       # dictionary to count undirected edges

    # The graph could be defined using a dictionary but nodes can be
    # added as subelements of the <network> element.  This allows
    # a node to be drawn differently from the rest

    # Let's find all the subelement nodes and store them in a dictionary
    nodes = {}
    positions = {}
    for node in element.findall('node'):
        handle = node.get('at', None)
        diagram.add_id(node, handle)
        handle = node.get('id')
        nodes[handle] = node

        # check for the position of this node 
        position = node.get('p', None)
        if position is not None:
            positions[handle] = un.valid_eval(position)

        # now check for edges
        edges = node.get('edges', None)
        if edges is not None:
            edges = un.valid_eval(edges)

            # tuples with a single element aren't returned as a list
            # this will be fixed but here's a workaround
            if isinstance(edges, np.ndarray):
                edges = list(edges)
            else:
                edges = [edges]

            # record each given edge
            for destination in edges:
                destination = str(destination)
                if destination == handle:
                    loop_count[handle] = loop_count.get(handle, 0) + 1
                    continue
                vertices = [handle, destination]

                leaving_edges = directed_edges.get(tuple(vertices), [])
                leaving_edges.append(None)
                directed_edges[tuple(vertices)] = leaving_edges

                vertices.sort()
                all_edges[tuple(vertices)] = all_edges.get(tuple(vertices), 0) + 1

    # Now we'll go through the structure given by the graph dictionary
    for node, edges in graph_dict.items():
        # have we seen this node already?
        if nodes.get(node, None) is None:
            nodes[node] = None
        for destination in edges:
            if destination == node:
                loop_count[node] = loop_count.get(node, 0) + 1
                continue
            vertices = [node, destination]

            leaving_edges = directed_edges.get(tuple(vertices), [])
            leaving_edges.append(None)
            directed_edges[tuple(vertices)] = leaving_edges
            
            vertices.sort()
            all_edges[tuple(vertices)] = all_edges.get(tuple(vertices), 0) + 1

    # finally, there may be <edge> subelements of <network> with decorations
    for edge in element.findall('edge'):
        endpoints = un.valid_eval(edge.get('vertices'))
        p = str(endpoints[0])
        q = str(endpoints[1])
        if p == q:
            loop_count[p] = loop_count.get(p, 0) + 1
            continue
        vertices = [p, q]

        placed = False
        leaving_edges = directed_edges.get(tuple(vertices), [])
        for i, e in enumerate(leaving_edges):
            if e is None:
                leaving_edges[i] = edge
                placed = True
                break
        if not placed and not directed:
            vertices = [q, p]
            leaving_edges = directed_edges.get(tuple(vertices), [])
            for i, e in enumerate(leaving_edges):
                if e is None:
                    leaving_edges[i] = edge
                    placed = True
                    break
        if not placed:
            vertices = [p, q]
            leaving_edges = directed_edges.get(tuple(vertices), [])
            leaving_edges.append(edge)
            vertices.sort()
            all_edges[tuple(vertices)] = all_edges.get(tuple(vertices), 0) + 1

        directed_edges[tuple(vertices)] = leaving_edges
        

    # Use networkx to layout the graph if we need to
    if len(positions) != len(nodes):
        # may need to use MultiGraph or MultiDiGraph
        # layout seems best with a plain Graph though
        G = nx.Graph()

        for key in nodes.keys():
            G.add_node(key)
        for edge, edges in directed_edges.items():
            for _ in range(len(edges)):
                G.add_edge(edge[0], edge[1])

        layout = element.get('layout', None)

        if layout is None or layout == 'spring':
            seed = int(element.get('seed', '1'))
            positions = nx.spring_layout(G, seed=seed)
        elif layout == 'bfs':
            start = element.get('start', None)
            if start is None:
                print('bfs network layout needs a starting node')
                return
            positions = nx.bfs_layout(G, start=start)
        elif layout == 'spectral':
            positions = nx.spectral_layout(G)

    # Now that we have the positions of the nodes, we will form the
    # graphical components.  First find the bounding box
    xvals = [p[0] for p in positions.values()]
    yvals = [p[1] for p in positions.values()]
    ll = np.array([min(xvals), min(yvals)])
    ur = np.array([max(xvals), max(yvals)])
    center = 0.5*(ll + ur)
    
    scale = float(element.get('scale', '0.8'))
    rotate = float(element.get('rotate', '0'))
    ctm = CTM.CTM()
    ctm.translate(*(-center))
    ctm.rotate(rotate)
    for key in positions.keys():
        positions[key] = ctm.transform(positions[key])

    xvals = [p[0] for p in positions.values()]
    yvals = [p[1] for p in positions.values()]
    ll = np.array([min(xvals), min(yvals)])
    ur = np.array([max(xvals), max(yvals)])
    center = 0.5*(ll+ur)

    ctm = CTM.CTM()
    ctm.translate(*(-center))
    for key in positions.keys():
        positions[key] = ctm.transform(positions[key])

    xvals = [p[0] for p in positions.values()]
    yvals = [p[1] for p in positions.values()]
    ll = np.array([min(xvals), min(yvals)])
    ur = np.array([max(xvals), max(yvals)])
    center = 0.5*(ll+ur)

    ll = 1/scale * ll
    ur = 1/scale * ur
    
    bbox_str = f"({ll[0]},{ll[1]},{ur[0]},{ur[1]})"

    # We're going to turn the <network> element into <coordinates>
    edge_stroke = element.get('edge-stroke', 'black')
    edge_thickness = element.get('edge-thickness', '2')
    edge_dash = element.get('edge-dash', 'none')
    node_fill = element.get('node-fill', 'darkorange')
    node_stroke = element.get('node-stroke', 'black')
    node_thickness = element.get('node-thickness', '1')
    node_style = element.get('node-style', 'circle')
    labels = element.get('labels', 'no') == 'yes'
    default_node_size = '10'
    if labels:
        default_node_size = '12'
    node_size = element.get('node-size', default_node_size)

    if diagram.output_format() == 'tactile':
        node_fill = 'white'
        node_stroke = 'black'

    element.clear()
    element.tag = 'coordinates'
    element.set('bbox', bbox_str)

    # We'll add groups for edges and then nodes. first the edges
    edge_group = ET.SubElement(element, 'group')
    edge_group.set('outline', 'tactile')
    
    spread = 15
    # remember that directed_edges is a dictionary:
    #   key = (p, q)
    #   value = all the edges from p to q
    for edge, edges in directed_edges.items():
        handle_0, handle_1 = edge
        endpoints = [handle_0, handle_1]
        endpoints.sort()
        edge = tuple(endpoints)
        y = (all_edges[edge] - 1)/2 * spread
#        for _ in range(len(edges)):
        for edge in edges:
            ctm = CTM.CTM()
            user_p0 = positions[handle_0]
            user_p1 = positions[handle_1]
            p0 = diagram.transform(user_p0)
            p1 = diagram.transform(user_p1)
            u = p1 - p0
            angle = math.atan2(u[1], u[0])
            length = math_util.length(u)
            ctm.translate(*p0)
            ctm.rotate(angle, units="rad")
            center = ctm.transform((length/2, y))
            c1 = ctm.transform((length/4, y))
            c2 = ctm.transform((3*length/4, y))
            center = diagram.inverse_transform(center)
            c1 = diagram.inverse_transform(c1)
            c2 = diagram.inverse_transform(c2)

            path = ET.SubElement(edge_group, 'path')
            if directed:
                path.set('mid-arrow', 'yes')

            path.set('start', util.pt2long_str(user_p0, spacer=","))
            if edge is None:
                path.set('stroke', edge_stroke)
                path.set('thickness', edge_thickness)
                path.set('dash', edge_dash)
            else:
                path.set('stroke', edge.get('stroke', edge_stroke))
                path.set('thickness', edge.get('thickness', edge_thickness))
                path.set('dash', edge.get('dash', edge_dash))

            curveto = ET.SubElement(path, 'quadratic-bezier')
            curveto.set('controls', ','.join(['('+util.pt2long_str(p, spacer=",")+')' for p in [c1, center]]))
            curveto = ET.SubElement(path, 'quadratic-bezier')
            curveto.set('controls', ','.join(['('+util.pt2long_str(p, spacer=",")+')' for p in [c2, user_p1]]))

            y -= spread

    node_group = ET.SubElement(element, 'group')
    node_group.set('outline', 'tactile')

    for handle, position in positions.items():
        node = nodes.get(handle, None)
        p = ET.SubElement(node_group, 'point')
        p.set('p', '(' + util.pt2long_str(position, spacer=',') + ')')
        p.set('size', node_size)

        if node is None:
            p.set('fill', node_fill)
            p.set('stroke', node_stroke)
            p.set('thickness', node_thickness)
            p.set('style', node_style)
        else:
            p.set('stroke', node.get('stroke', node_stroke))
            p.set('thickness', node.get('thickness', node_thickness))
            p.set('fill', node.get('fill', node_fill))
            p.set('style', node.get('style', node_style))

        if labels:
            label_text = handle
            if node is not None:
                label_text = node.get('label', label_text)
            l = ET.SubElement(node_group, 'label')
            math_element = ET.SubElement(l, 'm')
            math_element.text = label_text
            l.set('p', '(' + util.pt2long_str(position, spacer=',') + ')')
            l.set('alignment', 'center')
            l.set('offset', '(0,0)')
            l.set('clear-background', 'no')
            
    coordinates.coordinates(element, diagram, parent, outline_status)
#    diagram.save_network_coordinates(element, coords)



import numpy as np
import scipy
import scipy.special
import math

# introduce some useful mathematical operations
#   that are meant to be available to authors

def dot(u, v):
    return np.dot(np.array(u), np.array(v))

def length(u):
    return np.linalg.norm(np.array(u))

def normalize(u):
    return 1/length(u) * np.array(u)

def midpoint(u, v):
    return 0.5*(np.array(u) + np.array(v))

def angle(p, units = 'deg'):
    angle = math.atan2(p[1], p[0])
    if units == 'deg':
        return math.degrees(angle)
    return angle

def roll(array):
    array = np.array(array)
    return np.roll(array, 1, axis=0)

def choose(n,k):
    return int(scipy.special.binom(n,k))

def append(input, item):
    l = list(input)
    l.append(item)
    if isinstance(input, np.ndarray):
        l = np.array(l)
    return l

def chi_oo(a, b, t):
    if t > a and t < b:
        return 1
    return 0

def chi_oc(a, b, t):
    if t > a and t <= b:
        return 1
    return 0

def chi_co(a, b, t):
    if t >= a and t < b:
        return 1
    return 0

def chi_cc(a, b, t):
    if t >= a and t <= b:
        return 1
    return 0

def eulers_method(f, t0, y0, t1, N):
    h = (t1 - t0)/N
    if isinstance(y0, np.ndarray):
        points = [[t0, *y0]]
    else:
        points = [[t0, y0]]
    t = t0
    y = y0
    for _ in range(N):
        t += h
        y += f(t, y) * h
        if isinstance(y, np.ndarray):
            points.append([t, *y])
        else:
            points.append([t, y])
    return np.array(points)




import sys
import lxml.etree as ET
from . import annotations
from . import area
from . import clip
from . import circle
from . import coordinates
from . import definition
from . import diffeqs
from . import graph
from . import grid_axes
from . import group
from . import implicit
from . import label
from . import legend
from . import line
from . import network
from . import path
from . import parametric_curve
from . import point
from . import polygon
from . import rectangle
from . import riemann_sum
from . import repeat
from . import shape
from . import slope_field
from . import tangent_line
from . import vector

# this dictionary associates tags to a function that processes
#   elements having that tag

tag_dict = {
    'angle-marker': circle.angle,
    'annotations': annotations.annotations,
    'arc': circle.arc,
    'area-between-curves': area.area_between_curves,
    'area-under-curve': area.area_under_curve,
    'axes': grid_axes.axes,
    'caption': label.caption,
    'circle': circle.circle,
    'clip': clip.clip,
    'coordinates': coordinates.coordinates,
    'de-solve': diffeqs.de_solve,
    'definition': definition.definition,
    'define-shapes': shape.define,
    'derivative': definition.derivative,
    'ellipse': circle.ellipse,
    'graph': graph.graph,
    'grid': grid_axes.grid,
    'grid-axes': grid_axes.grid_axes,
    'group': group.group,
    'implicit-curve': implicit.implicit_curve,
    'label': label.label,
    'legend': legend.legend,
    'line': line.line,
    'network': network.network,
    'parametric-curve': parametric_curve.parametric_curve,
    'path': path.path,
    'plot-de-solution': diffeqs.plot_de_solution,
    'point': point.point,
    'polygon': polygon.polygon,
    'rectangle': rectangle.rectangle,
    'riemann-sum': riemann_sum.riemann_sum,
    'repeat': repeat.repeat,
    'shape': shape.shape,
    'slope-field': slope_field.slope_field,
    'tangent-line': tangent_line.tangent,
    'triangle': polygon.triangle,
    'vector': vector.vector
}

# apply the processing function based on the XML element's tag

def parse_element(element, diagram, root, outline_status = None):
    if element.tag is ET.Comment:
        return
    if path.is_path_tag(element.tag):
        print(f"A <{element.tag}> tag can only occur inside a <path>")
        return
    if label.is_label_tag(element.tag):
        print(f"A <{element.tag}> tag can only occur inside a <label>")
        return
    if grid_axes.is_axes_tag(element.tag):
        print(f"A <{element.tag}> tag can only occur inside a <axes> or <grid-axes>")
        return
        
    try:
        function = tag_dict[element.tag]
    except KeyError:
        print('Unknown element tag: ' + element.tag)
        sys.exit()
    function(element, diagram, root, outline_status)



import os
import shutil
import subprocess
from pathlib import Path
from . import core
import logging

# We're going to include some basic functions here so they can be
# called from an import

#log = logging.getLogger('ptxlogger')

def build(
        format,
        filename,
        publication=None,
        ignore_publication=True,
        standalone=False
):
    path = Path(filename)
    if path.suffix != '.xml':
        filename = str(path.parent / (path.stem + '.xml'))

#    log.info(f'Building from PreFigure source {filename}')

    # We're going to look for a publication, possibly in a parent directory
    # unless we're told to ignore any publication file
    if ignore_publication:
        publication = None
    else:
        if publication is None:
            pub_name = 'pf_publication.xml'
        else:
            pub_name = publication
        cwd = Path(os.getcwd())
        dirs = [cwd] + list(cwd.parents)
        for dir in dirs:
            pub = dir / pub_name
            if pub.exists():
                publication = pub
                break

    core.parse.parse(filename, format, publication, standalone)
    return filename


def pdf(
        format,
        filename,
        build_first=True,
        publication=None,
        ignore_publication=True,
        dpi=72
):
    build_path = None
    if build_first:
        filename = build(format,
                         filename,
                         publication=publication,
                         ignore_publication=ignore_publication)
        filename = Path(filename)
        build_path = filename.parent / 'output' / (filename.stem + '.svg')
    else:
        filename = Path(filename)
    
    if filename.suffix != '.svg':
        filename = filename.parent / (filename.name + '.svg')

    if build_path is None:
        filename_str = str(filename)
        for dir, dirs, files in os.walk(os.getcwd()):
            files = set(files)
            if filename_str in files:
                build_path = dir / filename
        if build_path is None:
#            log.debug(f'Unable to find {filename}')
            return

    dpi = str(dpi)
    executable = shutil.which('rsvg-convert')
    if executable is None:
#        log.debug('rsvg-convert is required to create PDFs.')
#        log.debug('See the installation instructions at https://prefigure.org')
        return
    
#    log.info(f'Converting {build_path} to PDF')
    output_file = build_path.parent / (build_path.stem + '.pdf')
    pdf_args = ['-a','-d',dpi,'-p',dpi,'-f','pdf','-o']
    pdf_args = ['rsvg-convert'] + pdf_args + [output_file,build_path]
    subprocess.run(pdf_args)

def png(
        format,
        filename,
        build_first=True,
        publication=None,
        ignore_publication=True
):
    build_path = None
    if build_first:
        filename = build(format,
                         filename,
                         publication=publication,
                         ignore_publication=ignore_publication)
        filename = Path(filename)
        build_path = filename.parent / 'output' / (filename.stem + '.svg')
    else:
        filename = Path(filename)
    
    if filename.suffix != '.svg':
        filename = filename.parent / (filename.stem + '.svg')

    if build_path is None:
        filename_str = str(filename)
        for dir, dirs, files in os.walk(os.getcwd()):
            files = set(files)
            if filename_str in files:
                build_path = dir / filename
        if build_path is None:
#            log.debug(f'Unable to find {filename}')
            return

#    log.info(f'Converting {build_path} to PDF')
    output_file = build_path.parent / (build_path.stem + '.png')

    import cairosvg
    cairosvg.svg2png(url=str(build_path), write_to=str(output_file))




import subprocess
import sys
import os
from pathlib import Path
import shutil


def main():
    prefig_root = Path(__file__).parent.parent
    destination = prefig_root / "core" / "mj_sre"

    shutil.copytree(
        prefig_root / "resources" / "js",
        destination,
        dirs_exist_ok = True
    )

    shutil.rmtree(destination / "node_modules", ignore_errors=True)

    # We need to change into mj_sre directory for installation
    # on Windows and linux
    wd = os.getcwd()
    os.chdir(destination)
    print(f"Installing MathJax libraries in {destination}")
    try:
        subprocess.run(["npm", "install"]) #, f"--prefix={destination}"])
    except:
        print("MathJax installation failed.  Is npm installed on your system?")
        sys.exit()
    os.chdir(wd)

        
if __name__ == "__main__":
    main()



import lxml.etree as ET
import numpy as np
import math
import copy
from . import user_namespace as un
from . import utilities
from . import grid_axes
from . import group
from . import math_utilities as math_util

# Add a graphical element for slope fields
def slope_field(element, diagram, parent, outline_status):
    if outline_status == 'finish_outline':
        finish_outline(element, diagram, parent)
        return

    f = un.valid_eval(element.get('function'))
    bbox = diagram.bbox()

    # We're going to turn this element into a group and add lines to it
    element.tag = "group"
    if element.get('outline', 'no') == 'yes':
        element.set('outline', 'always')

    # Now we'll construct a line with all the graphical information
    # and make copies of it
    line_template = ET.Element('line')

    if diagram.output_format() == 'tactile':
        line_template.set('stroke', 'black')
    else:
        line_template.set('stroke', element.get('stroke', 'blue'))
    line_template.set('thickness', element.get('thickness', '2'))
    if element.get('arrows', 'no') == 'yes':
        line_template.set('arrows', '1')

    if element.get('arrow-width', None) is not None:
        line_template.set('arrow-width', element.get('arrow-width'))
    if element.get('arrow-angles', None) is not None:
        line_template.set('arrow-angles', element.get('arrow-angles'))

    # Now we'll construct each of the lines in the slope field
    system = element.get('system', None) == 'yes'
    spacings = element.get('spacings', None)
    if spacings is not None:
        spacings = un.valid_eval(spacings)
        rx, ry = spacings
    else:   
        rx = grid_axes.find_gridspacing((bbox[0], bbox[2]))
        ry = grid_axes.find_gridspacing((bbox[1], bbox[3]))

    x = rx[0]
    while x <= rx[2]:
        y = ry[0]
        while y <= ry[2]:
            line = copy.deepcopy(line_template)
            if system:
                change = f(0, [x,y])
                if math_util.length(change) > 1e-05:
                    element.append(line)
                if abs(change[0]) < 1e-08:
                    dx = 0
                    dy = ry[1]/4
                    if change[1] < 0:
                        dy *= -1
                else:
                    slope = change[1]/change[0]
                    dx = rx[1]/4
                    dy = slope*dx
                    if abs(dy) > ry[1]/4:
                        dy = ry[1]/4
                        dx = dy/slope
                    if change[0] * dx < 0:
                        dx *= -1
                        dy *= -1
            else:
                slope = f(x,y)
                dx = rx[1]/4
                dy = slope*dx
                if abs(dy) > ry[1]/4:
                    dy = ry[1]/4
                    dx = dy/slope
                if dx < 0:
                    dx *= -1
                    dy *= -1
                element.append(line)
            x0 = x - dx
            x1 = x + dx
            y0 = y - dy
            y1 = y + dy
            line.set('p1', utilities.pt2long_str((x0,y0), spacer=','))
            line.set('p2', utilities.pt2long_str((x1,y1), spacer=','))
            y += ry[1]
        x += rx[1]

    group.group(element, diagram, parent, outline_status)




from . import user_namespace as un

# allows authors to define mathematical quantities
#    to be used in the diagram

def definition(element, diagram, parent, outline_status):
    substitution = element.get('substitution', 'yes') == 'yes'
    un.define(element.text, substitution)

    id_suffix = element.get('id-suffix')
    if id_suffix is not None:  # this definition is part of a repeat
        diagram.push_id_suffix('-' + id_suffix) 
        diagram.parse(element, parent, outline_status)
        diagram.pop_id_suffix()

def derivative(element, diagram, parent, outline_status):
    f = un.valid_eval(element.get('function'))
    name = element.get('name')
    un.derivative(f, name)



import os
import sys
import re
import math
import inspect
from pathlib import Path
import numpy as np
import lxml.etree as ET
from . import utilities as util
from . import CTM
from . import user_namespace as un
import tempfile
try:
    import cairo
except:
    print('Error importing Python package cairo, which is required for non-mathemaical labels.')
    print('See the PreFigure installation instructions at https://prefigure.org')

# These are tags that can occur in a label
label_tags = {'it', 'b', 'newline'}

def is_label_tag(tag):
    return tag in label_tags

# Labels will be handled here.
# An important point is how a label will be aligned relative to its anchor.
# These dictionaries define alignments on a 3x3 grid with the anchor at the center.
alignment_displacement = {
    'southeast': [0, 0],
    'east': [0, 0.5],
    'northeast': [0, 1],
    'north': [-0.5, 1],
    'northwest': [-1, 1],
    'west': [-1, 0.5],
    'southwest': [-1, 0],
    'south': [-0.5, 0],
    'center': [-0.5, 0.5],
    'se': [0, 0],
    'e':  [0, 0.5],
    'ne': [0, 1],
    'n':  [-0.5, 1],
    'nw': [-1, 1],
    'w':  [-1, 0.5],
    'sw': [-1, 0],
    's':  [-0.5, 0],
    'c':  [-0.5, 0.5],
    'xaxis-label': [-0.5, 0],
    'ha': [-0.5, 0],  # horizontal axis label
    'va': [-1, 0.5],  # vertical axis label
    'xl': [-1, 1]
}

braille_displacement = {
    'southeast': [0, -1],
    'east': [0, -0.5],
    'northeast': [0, 0],
    'north': [-0.5, 0],
    'northwest': [-1, 0],
    'west': [-1, -0.5],
    'southwest': [-1, -1],
    'south': [-0.5, -1],
    'center': [-0.5, -0.5],
    'se': [0, -1],
    'e':  [0, -0.5],
    'ne': [0, 0],
    'n':  [-0.5, 0],
    'nw': [-1, 0],
    'w':  [-1, -0.5],
    'sw': [-1, -1],
    's':  [-0.5, -1],
    'c':  [-0.5, -0.5],
    'xaxis-label': [0, -1],
    'ha': [0, -1],     # horizontal axis label
    'hat': [0, 0],    # top horizontal axis label
    'va': [-1, -0.5],  # vertical axis label
    'va': [0, -0.5],   # right vertical axis label
    'xl': [-1, 0]
}

# For labels on arrows
alignment_circle = [
    'east', 'northeast', 'north', 'northwest',
    'west', 'southwest', 'south', 'southeast'
]

def get_alignment_from_direction(direction):
    direction_angle = math.degrees(math.atan2(direction[1], direction[0]))
    align = round(direction_angle/45) % 8
    return alignment_circle[align]

nemeth_on =  '⠸⠩ '
nemeth_off = '⠸⠱ '

# We use pycairo to measure the dimensions of svg text labels
# so we need a cairo context.  This is not needed for tactiles diagrams

try:
    surface = cairo.SVGSurface(None, 200, 200)
    context = cairo.Context(surface)
    context.select_font_face('sans')
    font_size = 14
    context.set_font_size(font_size)
    cairo_context = context
except:
    # If cairo is not installed, we've already given a warning so we'll keep going
    pass

# Now we'll place a label into a diagram.  Labels are created by
# mathjax so we're going to put all the labels into an HTML file
# to be processed together.  As a result, labels will be added to
# the diagram after all the other components have been processed.

def label(element, diagram, parent, outline_status = None):
    if outline_status == 'add_outline':  # we're not ready for labels
        return

    # Define a group to hold the label.  
    group = ET.Element('g')
    diagram.add_label(element, group)
    diagram.add_id(element, element.get('id'))
    group.set('id', element.get('id'))

    # For non-tactile output, add a group to the diagram.
    # We'll return and insert the label into the group later
    # after everything else has been processed
    if diagram.output_format() != 'tactile':
        parent.append(group)

    # We first go pull out the <m> tags and write them into
    # an HTML file to be processed by MathJax
    # We also want to know if the label is a single lower-case
    # letter so that we can add a letter indicator
    text = element.text
    if text is None:
        text = ''
    else: element.text = evaluate_text(text)
    plain_text = text
    for math in element.findall('m'):
        diagram.add_id(math)
        math_id = math.get('id')
        math.text = evaluate_text(math.text).strip()
        math_text = '\({}\)'.format(math.text)

        # add the label's text to the HTML tree
        div = ET.SubElement(diagram.label_html(), 'div')
        div.set('id', math_id)
        div.text = math_text

        text += math_text
        plain_text += str(math_text)
        if math.tail is not None:
            math.tail = evaluate_text(math.tail)
            text += math.tail
            plain_text += str(math.tail)
    text = text.strip()

    align = util.get_attr(element, 'alignment', 'c')
    if align.startswith('2') or align == 'e':
        align = 'east'

    element.set('alignment', align)
    if element.get('anchor', None) is not None:
        element.set('p', element.get('anchor'))
    element.set('p', util.get_attr(element, 'p', '[0,0]'))

    # if we're making a tactile diagram and the text is a single
    # lower-case letter, we'll add a letter indicator in front
    if diagram.output_format() == 'tactile' and len(plain_text) == 1:
        char_distance = ord(plain_text[0]) - ord('a')
        if char_distance >= 0 and char_distance < 26:
            element.set('add-letter-indicator', 'yes')

# Allow substitutions from the user namespace
def evaluate_text(text):
    tokens = re.split(r"(\${[^}]*})", text)
    return ''.join([str(un.valid_eval(token[2:-1])) if token.startswith('${') else token for token in tokens])

def place_labels(diagram, filename, root, label_group_dict, label_html_tree):
    # if there are no labels, there's nothing to do
    if len(label_group_dict) == 0:
        return

    # prepare the MathJax command
    output_format = diagram.output_format()
    filename = filename[:-4]
    basename = os.path.basename(filename)
    working_dir = tempfile.TemporaryDirectory()
    mj_input = os.path.join(working_dir.name, basename) + '-labels.html'
    mj_output = os.path.join(working_dir.name, basename) + '-' + output_format + '.html'

    # write the HTML file
    with ET.xmlfile(mj_input, encoding='utf-8') as xf:
        xf.write(label_html_tree, pretty_print=True)

    options = ''
    if output_format == 'tactile':
        format = 'braille'
    else:
        options = '--svgenhanced --depth deep'
        format = 'svg'

    # have MathJax process the HTML file and load the resulting
    # SVG labels into label_tree 
    path = Path(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))))
    mj_dir = path.absolute() / 'mj_sre'
    mj_dir_str = str(mj_dir)

    if not (mj_dir / 'mj-sre-page.js').exists():
        from .. import scripts
        scripts.install_mj.main()
            

    mj_command = 'node {}/mj-sre-page.js --{} {} {} > {}'.format(mj_dir_str, format, options, mj_input, mj_output)

    os.system(mj_command)
    label_tree = ET.parse(mj_output)

    # for braille output, we'll create a group to hold all the labels
    # and their clear backgrounds and add it at the end of the diagram
    if diagram.output_format() == 'tactile':
        try:
            global louis
            import louis
        except:
            print('Failed to import louis so we cannot make braille labels')
            print('See the installation instructions at https://prefigure.org')
            print('The rest of the diagram will still be built.')
            return
        background_group = ET.SubElement(root, 'g')
        background_group.set('id', 'background-group')
        braille_group = ET.SubElement(root, 'g')
        braille_group.set('id', 'braille-group')

    # now go through each label and place them in the diagram
    for label, group_ctm in label_group_dict.items():
        group, ctm = group_ctm

        if diagram.output_format() == 'tactile':
            position_braille_label(label, diagram, ctm, background_group, 
                                   braille_group, label_tree)
        else:
            position_svg_label(label, diagram, ctm, group, label_tree)

    working_dir.cleanup()

# use this to retrieve elements from the mathjax output
#        div = label_tree.xpath("//html/body/div[@id = '{}']".format(id))[0]

def position_braille_label(element, diagram, ctm, 
                           background_group, braille_group, label_tree):
    group = ET.SubElement(braille_group, 'g')
    group.set('id', element.get('id'))

    # Determine the anchor point p and then adjust it using
    # the alignment and offset
    if element.get('user-coords', 'yes') == 'yes':
        p = ctm.transform(un.valid_eval(element.get('p')))
    else:
        p = un.valid_eval(element.get('p'))
    alignment = util.get_attr(element, 'alignment', 'center')
    displacement = braille_displacement[alignment]

    # TODO:  improve automatic tactile offsets
    offset = util.get_attr(element, 'abs-offset', 'none')
    if offset == 'none':
        offset = [8*(displacement[0] + 0.5), 8*(displacement[1]+0.5)]
    else:
        offset = un.valid_eval(offset)

    offset = [o + 6*np.sign(o) for o in offset]
    if displacement[0] == 0:
        offset[0] += 6
    if displacement[1] == -1:
        offset[1] -= 6
    if alignment == 'n' or alignment == 'north':
        offset[1] += 5

    # 'gap' is the distance, in points, between two braille dots,
    # 1/20th of an inch
    gap = 3.6
    if alignment == 'ha':
        offset = [-4*gap, -30]
    if alignment == 'hat':
        offset = [-4*gap, 30]
    if alignment == 'va':
        offset = [-30, 0]
    if alignment == 'var':
        offset = [30, 0]
    if alignment == 'xl':  
        offset = [-10, 12]

    if element.get('offset', None) is not None:
        relative_offset = un.valid_eval(element.get('offset'))
        offset = [offset[0] + relative_offset[0],
                  offset[1] + relative_offset[1]]

    p[0] += offset[0]
    p[1] -= offset[1]

    # let's assemble the different pieces
    row = [[element.text, 'plain']]
    text_elements = [row]
    
    for el in element:
        if el.tag == 'newline':
            row = []
            text_elements.append(row)
        if el.tag == 'm':
            row.append(el)
        if el.tag == 'it':
            row.append([el.text, 'it'])
            for child in el:
                if child.tag != 'b':
                    print(f"<{child.tag}> is not allowed inside a <it>")
                    continue
                row.append([child.text, 'it'])
                row.append([child.tail, 'it'])
        
        if el.tag == 'b':
            row.append([el.text, 'b'])
            for child in el:
                if child.tag != 'it':
                    print(f"<{child.tag}> is not allowed inside a <b>")
                    continue
                row.append([child.text, 'b'])
                row.append([child.tail, 'b'])
        row.append([el.tail, 'plain'])

    # Let's make another pass through the elements removing
    # empty text and adding whitespace
    for num, row in enumerate(text_elements):
        # remove empty text
        new_row = []
        for el in row:
            if isinstance(el, list): # is this a text
                text = el[0]
                if text is not None:
                    text = text.strip()
                    if len(text) > 0:
                        if (
                                len(new_row) > 0 and
                                new_row[-1][1] == el[1]
                        ):
                            new_row[-1][0] += ' ' + text
                        else:
                            new_row.append([text, el[1]])
            else: # otherwise it's an <m>
                new_row.append(el)
        text_elements[num] = new_row

    typeform_dict = {'plain':0, 'it':1, 'b':4}
    space = louis.translateString(
        ["braille-patterns.cti", "en-us-g2.ctb"],
        ' ',
        typeform=[0]
    )
    # translate braille strings not in an <m>
    for num, row in enumerate(text_elements):
        row_text = ''
        while len(row) > 0:
            el = row.pop(0)
            if isinstance(el, list):
                text = el[0]
                if len(row) > 0:
                    text += ' '
                typeform = [typeform_dict[el[1]]] * len(text)
                braille_text = louis.translateString(
                    ["braille-patterns.cti", "en-us-g2.ctb"],
                    text,
                    typeform=typeform
                )
                row_text += braille_text
            else:
                m_tag_id = el.get('id')

                div = label_tree.xpath("//html/body/div[@id = '{}']".format(m_tag_id))[0]

                try:
                    insert = div.xpath('mjx-data/mjx-braille')[0]
                except IndexError:
                    print('Error in processing label, possibly a LaTeX error: ' + div.text)
                    sys.exit()
                row_text += insert.text
                if len(row) > 0:
                    row_text += space
        text_elements[num] = row_text

    interline = 28.8  # 0.4 inches
    width = 5 * gap * max([len(row) for row in text_elements])
    height = 5 * gap + interline * (len(text_elements) - 1)

    diagram.register_label_dims(element, (width, height))
    
    p[0] += width*displacement[0]
    p[1] -= height*displacement[1]

    # snap the point onto the 20dpi embossing grid
    p = [3.6 * round(c/3.6) for c in p]

    tform = CTM.translatestr(*p)
    group.set('transform', tform)

    # add a white background behind the label
    bg_margin = 9  # 1/8th of a inch
    rect = ET.SubElement(background_group, 'rect')
    rect.set('id', element.get('id') + '-background')
    rect.set('x', util.float2str(p[0]-bg_margin))
    rect.set('y', util.float2str(p[1]-height-bg_margin))
    #rect.set('x', util.float2str(-bg_margin))
    #rect.set('y', util.float2str(-height-bg_margin))
    rect.set('width', util.float2str(width+2*bg_margin))
    rect.set('height', util.float2str(height+2*bg_margin))
    rect.set('stroke', 'none')
    rect.set('fill', 'white')

    # show the lower-left corner point for debugging
    circle = ET.Element('circle')
    circle.set('cx', util.float2str(p[0]))
    circle.set('cy', util.float2str(p[1]))
    circle.set('r', '3')
    circle.set('fill', '#00f')
#    group.append(circle)

    # Now add the labels
    justify = element.get('justify', 'center')
    '''
    x = p[0] 
    y = p[1] - height + 5*gap
    '''
    x = 0
    y = -height + 5*gap
    for el in text_elements:
        text_element = ET.SubElement(group, 'text')
        x_line = x
        if justify == 'right':
            x_line = x + width - 5*gap*len(el)
        if justify == 'center':
            x_line = x+ (width - 5*gap*len(el))/2
            x_line = gap * round(x_line/gap)
        text_element.set('x', util.float2str(x_line))
        text_element.set('y', util.float2str(y))
        text_element.text = el
        text_element.set('font-family', "Braille29")
        text_element.set('font-size', "29px")
        y += interline


def position_svg_label(element, diagram, ctm, group, label_tree):
    # We're going to put everything inside a group
    label_group = ET.Element('g')

    # Determine the anchor point p and then adjust it using
    # the alignment and offset
    # TODO:  improve auto offsets
    if element.get('user-coords', 'yes') == 'yes':
        p = ctm.transform(un.valid_eval(element.get('p')))
    else:
        p = un.valid_eval(element.get('p'))
    alignment = util.get_attr(element, 'alignment', 'center')
    displacement = alignment_displacement[alignment]

    offset = util.get_attr(element, 'abs-offset', 'none')
    if offset == 'none':
        offset = [8*(displacement[0] + 0.5), 8*(displacement[1]-0.5)]
    else:
        offset = un.valid_eval(offset)

    if element.get('offset', None) is not None:
        relative_offset = un.valid_eval(element.get('offset'))
        offset = [offset[0] + relative_offset[0],
                  offset[1] + relative_offset[1]]


    # A label can have rows consisting of different components
    # comprised of text, with italics and bold, and <m> tags.
    # Our first task is to extract all of these components and measure them

    std_font_face = ['sans', 14, 'normal', 'normal']
    it_font_face =  ['sans', 14, 'italic', 'normal']
    b_font_face  =  ['sans', 14, 'normal', 'bold']
    it_b_font_face  =  ['sans', 14, 'italic', 'bold']

    label = element

    # We first pass through all the text and elements in the label
    # extracting the text and the font information

    row = [(label.text, std_font_face)]
    text_elements = [row]
    
    for el in label:
        if el.tag == 'newline':
            row = []
            text_elements.append(row)
        if el.tag == 'm':
            row.append(el)
        if el.tag == 'it':
            row.append((el.text, it_font_face))
            for child in el:
                if child.tag != 'b':
                    print(f"<{child.tag}> is not allowed inside a <it>")
                    continue
                row.append((child.text, it_b_font_face))
                row.append((child.tail, it_font_face))
        
        if el.tag == 'b':
            row.append((el.text, b_font_face))
            for child in el:
                if child.tag != 'it':
                    print(f"<{child.tag}> is not allowed inside a <b>")
                    continue
                row.append((child.text, it_b_font_face))
                row.append((child.tail, b_font_face))
        row.append((el.tail, std_font_face))

    # Let's make another pass through the elements removing
    # empty text and adding whitespace
    for num, row in enumerate(text_elements):
        new_row = []
        begin_space = False
        for element in row:
            if isinstance(element, tuple): # is this a text
                text = element[0]
                if text is not None:
                    text = text.strip()
                    if len(text) > 0:
                        new_row.append(mk_text_element(text,
                                                       element[1],
                                                       label_group))
                
            else: # otherwise it's an <m>
                new_row.append(mk_m_element(element,
                                            label_tree,
                                            label_group))
                begin_space = True
        text_elements[num] = new_row

    # let's go through each row and find the dimensions
    cairo_context.select_font_face('sans',
                                   cairo.FontSlant.NORMAL,
                                   cairo.FontWeight.NORMAL
                                   )
    space = cairo_context.text_extents(' ')[4]
    interline = un.valid_eval(label.get('interline', '3'))
    text_dimensions = []
    for num, row in enumerate(text_elements):
        width = 0
        above = [0]
        below = [0]
        for el in row:
            width += el[1]
            above.append(el[2])
            below.append(el[3])
        above = max(above)
        below = max(below)
        height = above + below
        width += (len(row)-1) * space
        if num == len(text_elements) - 1:
            interline = 0
        text_dimensions.append([width,
                                height+interline,
                                above,
                                below])

    # let's find the dimension of the bounding box
    width = max([row[0] for row in text_dimensions])
    height = sum([row[1] for row in text_dimensions])

    diagram.register_label_dims(label, (width, height))

    # finally, we set the location of each subelement
    justify = label.get('justify', 'center')
    y_location = 0
    for row, dims in zip(text_elements, text_dimensions):
        row_width = dims[0]
        row_above = dims[2]
        x_location = 0
        if justify == 'center':
            x_location = (width - row_width)/2
        if justify == 'right':
            x_location = width - row_width
        for el in row:
            component = el[0]
            component.set('x', util.float2str(x_location))
            x_location += el[1] + space
            if component.tag == 'text':
                component.set('y',
                              util.float2str(y_location+row_above)
                              )
            else:
                component.set('y',
                              util.float2str(y_location+
                                             row_above-
                                             el[2])
                              )
        y_location += dims[1]
                


    # Now we find the coordinate transform for the label_group
    tform = CTM.translatestr(p[0] + offset[0], p[1] - offset[1])

    sc = float(label.get('scale', '1'))
    if sc != 1:
        tform = tform + ' ' + CTM.scalestr(sc, sc)

    rot = label.get('rotate', None)
    if rot is not None:
        tform = tform + ' ' + CTM.rotatestr(float(rot))
    tform = tform + ' ' + CTM.translatestr(width*displacement[0], 
                                           -height*displacement[1])

    group.set('transform', tform)

    # add a white rectangle behind the label, if requested
    if label.get('clear-background', 'no') == 'yes':
        bg_margin = int(label.get('background-margin', '6'))
        rect = ET.SubElement(group, 'rect')
        rect.set('x', str(-bg_margin))
        rect.set('y', str(-bg_margin))
        rect.set('width', str(width+2*bg_margin))
        rect.set('height', str(height+2*bg_margin))
        rect.set('stroke', 'none')
        rect.set('fill', 'white')

    group.append(label_group)
    diagram.add_id(label_group, label.get('expr'))
    group.set('type', 'label')


def mk_text_element(text_str, font_face, label_group):
    text_el = ET.SubElement(label_group, 'text')
    text_el.text = text_str
    text_el.set('font-family', font_face[0])
    text_el.set('font-size', str(font_face[1]))
    if font_face[2] == 'italic':
        text_el.set('font-style', 'italic')
    if font_face[3] == 'bold':
        text_el.set('font-weight', 'bold')

    if font_face[2] == 'italic':
        font_slant = cairo.FontSlant.ITALIC
    else:
        font_slant = cairo.FontSlant.NORMAL
    if font_face[3] == 'bold':
        font_weight = cairo.FontWeight.BOLD
    else:
        font_weight = cairo.FontWeight.NORMAL
    cairo_context.select_font_face(font_face[0],
                                   font_slant,
                                   font_weight)
    cairo_context.set_font_size(font_face[1])
    extents = cairo_context.text_extents(text_str)
    y_bearing = extents[1]
    t_height  = extents[3]
    xadvance  = extents[4]
    return [text_el, xadvance, -y_bearing, t_height+y_bearing]


def mk_m_element(m_tag, label_tree, label_group):
    m_tag_id = m_tag.get('id')
    div = label_tree.xpath("//html/body/div[@id = '{}']".format(m_tag_id))[0]
    ns = {'svg': 'http://www.w3.org/2000/svg'}
    try:
        insert = div.xpath('mjx-data/mjx-container/svg:svg',
                           namespaces=ns)[0]
    except IndexError:
        print('Error in processing label, possibly a LaTeX error: ' + div.text)
        sys.exit()

    # Express dimensions in px for rsvg-convert
    dim_dict = {}
    for attr in ['style', 'width', 'height']:
        fields = insert.get(attr).split()
        dimension = fields[-1]
        index = dimension.find('ex')
        dimension = float(dimension[:index]) * 8
        dim_dict[attr] = dimension
        pts = '{0:.3f}px'.format(dimension)
        fields[-1] = pts
        insert.set(attr, ' '.join(fields))
    label_group.append(insert)   

    width = dim_dict['width']

    above = dim_dict['height'] + dim_dict['style']
    below = -dim_dict['style']
    return [insert, width, above, below]


# add a caption to a tactile diagram in the upper-left corner
#   e.g. "Figure 2.3.4"
def caption(element, diagram, parent, outline_status):
    if diagram.output_format() != "tactile":  
        return
    element.tag = 'label'
    element.set('alignment', 'ne')
    element.set('offset', '(0,10)')
    box = diagram.bbox()
    element.set('p', util.pt2str((box[0], box[3]), spacer=","))
    label(element, diagram, parent)

    
